from __future__ import annotations

import base64
import mimetypes
from pathlib import Path
from typing import Dict, cast

from pydantic import Field

from kubernetes_dynamic.models.common import V1ObjectMeta

from .resource_item import ResourceItem


class V1ConfigMap(ResourceItem):
    binaryData: Dict[str, str] = Field(default_factory=dict)
    data: Dict[str, str] = Field(default_factory=dict)
    immutable: bool = False

    @staticmethod
    def _path_to_data_item(file_path: str | Path):
        file_path = Path(file_path)
        mimetype = mimetypes.guess_type(str(file_path))[0]
        item = {"data": {}, "binaryData": {}}
        if mimetype and mimetype.startswith("text"):
            with file_path.open("rt") as f:
                item["data"] = {file_path.name: f.read()}
        else:
            with file_path.open("rb") as f:
                item["binaryData"] = {file_path.name: base64.b64encode(f.read()).decode()}
        return item

    @classmethod
    def from_path(cls, name: str, namespace: str, file_path: str | Path):
        """Instantiate a configmap from file."""
        file_path = Path(file_path)
        data = {}
        binary_data = {}
        items = [file_path] if not file_path.is_dir() else [path for path in file_path.iterdir() if not path.is_dir()]
        for item in items:
            item_data = V1ConfigMap._path_to_data_item(item)
            data.update(item_data["data"])
            binary_data.update(item_data["binaryData"])
        configmap = cls(
            metadata=cast(
                V1ObjectMeta,
                {
                    "name": name,
                    "namespace": namespace,
                },
            ),
            apiVersion="v1",
            kind="ConfigMap",
            data=data,
            binaryData=binary_data,
        )
        return configmap



from kubernetes_dynamic.models.ingress import V1Ingress


def test_ingress_init():
    pod = V1Ingress.parse_obj(dict(metadata={"name": "my_ingress"}))
    assert pod.metadata.name == "my_ingress"


def test_ingress_get_default_host(mock_client):
    mock_client.ingresses.get.return_value = [
        V1Ingress.parse_obj(dict(metadata={"name": "my_namespace"}, spec={"rules": [{"host": "hostname"}]}))
    ]
    assert V1Ingress.get_default_host() == "https://hostname"


def test_ingress_get_default_host_empty(mock_client):
    mock_client.ingresses.get.return_value = []
    assert V1Ingress.get_default_host() is None



from __future__ import annotations

from pathlib import Path
from typing import Dict, Optional

from . import _kubernetes
import pydantic


class MutedSecretsSettingsSource:
    """Disable warning when secrets dir does not exist."""

    def __init__(self, source: pydantic.env_settings.SecretsSettingsSource):
        self.source = source

    def __call__(self, settings: pydantic.BaseSettings):
        secrets: Dict[str, Optional[str]] = {}

        if self.source.secrets_dir is None:
            return secrets

        secrets_path = Path(self.source.secrets_dir).expanduser()

        if not secrets_path.exists():
            return secrets
        return self.source(settings)


class K8sConfig(pydantic.BaseSettings):
    """Kubernetes configuration options."""

    configuration: Optional[_kubernetes.Configuration] = None
    context: str = pydantic.Field(default="", env="KUBE_CONTEXT")
    namespace: str = pydantic.Field(default="default", env=["NAMESPACE", "KUBE_NAMESPACE"])
    connection_timeout: int = pydantic.Field(default=10, env="KUBERNETES_CONNECTION_TIMEOUT")
    read_timeout: int = pydantic.Field(default=30, env="KUBERNETES_READ_TIMEOUT")

    class Config:  # noqa: D106
        arbitrary_types_allowed = True
        secrets_dir = "/run/secrets/kubernetes.io/serviceaccount"

        @classmethod
        def customise_sources(
            cls,
            init_settings,
            env_settings,
            file_secret_settings,
        ):
            """Configure settings priority in return order."""
            return init_settings, env_settings, MutedSecretsSettingsSource(file_secret_settings)



from unittest.mock import MagicMock

import pytest
from pytest_mock import MockerFixture

from kubernetes_dynamic.models.configmap import V1ConfigMap


def test_configmap_init():
    pod = V1ConfigMap.parse_obj(dict(metadata={"name": "my_configmap"}))
    assert pod.metadata.name == "my_configmap"


@pytest.mark.parametrize(
    "mimetype, read, data, binary_data",
    [
        ("text", "data", {"file_path": "data"}, {}),
        (None, b"\a4", {}, {"file_path": "BzQ="}),
    ],
)
def test_configmap_from_path_file(mimetype, read, data, binary_data, mocker: MockerFixture):
    path_mock = mocker.patch("kubernetes_dynamic.models.configmap.Path").return_value
    mimetypes_mock = mocker.patch("kubernetes_dynamic.models.configmap.mimetypes")
    path_mock.is_dir.return_value = False
    path_mock.open.return_value.__enter__.return_value.read.return_value = read
    path_mock.name = "file_path"
    mimetypes_mock.guess_type.return_value = (mimetype, None)
    configmap = V1ConfigMap.from_path("cm-name", "cm-namespace", "file_path")
    assert configmap.data == data
    assert configmap.binaryData == binary_data


def test_configmap_from_path_dir(mocker: MockerFixture):
    path_mock = mocker.patch("kubernetes_dynamic.models.configmap.Path").return_value
    path_data_mock = mocker.patch("kubernetes_dynamic.models.configmap.V1ConfigMap._path_to_data_item")
    path_mock.is_dir.return_value = True
    file_1 = MagicMock(is_dir=lambda: False, **{"name": "file_1"})
    file_2 = MagicMock(is_dir=lambda: False, **{"name": "file_2"})
    file_3 = MagicMock(is_dir=lambda: True, **{"name": "file_3"})
    path_mock.iterdir.return_value = [file_1, file_2, file_3]
    path_data_mock.side_effect = [
        {"data": {"file_1": "data"}, "binaryData": {}},
        {"data": {}, "binaryData": {"file_2": "BzQ="}},
    ]
    configmap = V1ConfigMap.from_path("cm-name", "cm-namespace", "file_path")
    assert configmap.data == {"file_1": "data"}
    assert configmap.binaryData == {"file_2": "BzQ="}



__all__ = [
    "ApiException",
    "DynamicApiError",
    "ResourceNotFoundError",
    "ResourceNotUniqueError",
    "KubernetesValidateMissing",
    "BadRequestError",
    "UnauthorizedError",
    "ForbiddenError",
    "NotFoundError",
    "MethodNotAllowedError",
    "ConflictError",
    "GoneError",
    "UnprocessibleEntityError",
    "TooManyRequestsError",
    "InternalServerError",
    "ServiceUnavailableError",
    "ServerTimeoutError",
    "ConfigException",
    "api_exception",
]


from kubernetes_dynamic.kube.exceptions import (
    BadRequestError,
    ConfigException,
    ConflictError,
    DynamicApiError,
    ForbiddenError,
    GoneError,
    InternalServerError,
    KubernetesValidateMissing,
    MethodNotAllowedError,
    NotFoundError,
    ResourceNotFoundError,
    ResourceNotUniqueError,
    ServerTimeoutError,
    ServiceUnavailableError,
    TooManyRequestsError,
    UnauthorizedError,
    UnprocessibleEntityError,
    api_exception,
)
from kubernetes_dynamic.openapi_client.exceptions import ApiException


class InvalidParameter(Exception):
    pass


class EventTimeoutError(TimeoutError):
    """Used when a waiting for an event times out."""

    def __init__(self, *args, last):
        super().__init__(*args)
        self.last = last



from unittest.mock import MagicMock

import pydantic
import pytest

from kubernetes_dynamic.client import K8sClient
from kubernetes_dynamic.events import Event
from kubernetes_dynamic.kube.exceptions import NotFoundError
from kubernetes_dynamic.kube.resource_api import ResourceApi
from kubernetes_dynamic.models.pod import V1Pod
from kubernetes_dynamic.models.resource_value import ResourceValue


@pytest.fixture
def fake_api(mock_client: K8sClient):
    return ResourceApi(name="pods", kind="Pod", api_version="v1", prefix="", namespaced=True, client=mock_client)


def fake_pod():
    return V1Pod.parse_obj({"kind": "Pod", "apiVersion": "v1", "metadata": {"name": "pod-name"}})


def test_k8s_resource_get(fake_api: ResourceApi, mock_client: K8sClient):
    """Test get method."""
    mock_client.request.return_value = fake_pod()
    assert fake_api.get("name", "namespace") == fake_pod()

    mock_client.request.side_effect = NotFoundError(MagicMock())
    assert fake_api.get("name", "namespace") is None


def test_find(fake_api: ResourceApi, mock_client: K8sClient):
    """Test generic find method."""
    pods = [fake_pod()]
    mock_client.request.return_value = pods
    assert fake_api.find("pod-.*", "namespace") == pods

    mock_client.request.side_effect = [pods]
    assert fake_api.find("something.*", "namespace") == []


def test_k8s_client_get_clusterwide(fake_api: ResourceApi, mock_client: K8sClient):
    mock_client.request.return_value = fake_pod()
    fake_api.namespaced = False
    assert fake_api.get("name", "namespace") == fake_pod()


def test_k8s_client_get_kwarg_namespace(fake_api: ResourceApi, mock_client: K8sClient):
    mock_client.request.return_value = fake_pod()
    assert fake_api.get("name", namespace="namespace") == fake_pod()


def test_k8s_client_get_404(fake_api: ResourceApi, mock_client: K8sClient):
    mock_client.request.side_effect = (NotFoundError(fake_pod()),)
    assert fake_api.get("name", "namespace") is None


def test_k8s_client_create(fake_api: ResourceApi, mock_client: K8sClient):
    mock_client.request.return_value = fake_pod()
    assert fake_api.create(MagicMock(), "namespace") == fake_pod()


def test_k8s_client_delete(fake_api: ResourceApi, mock_client: K8sClient):
    mock_client.request.return_value = fake_pod()
    assert (
        fake_api.delete("name", "namespace", {}, label_selector="label_selector", field_selector="field_selector")
        == fake_pod()
    )


def test_k8s_client_replace(fake_api: ResourceApi, mock_client: K8sClient):
    mock_client.request.return_value = fake_pod()
    assert (
        fake_api.replace(
            MagicMock(),
            "name",
            "namespace",
        )
        == fake_pod()
    )


def test_k8s_client_patch(fake_api: ResourceApi, mock_client: K8sClient):
    mock_client.request.return_value = fake_pod()
    assert (
        fake_api.patch(
            MagicMock(),
            "name",
            "namespace",
        )
        == fake_pod()
    )


def test_k8s_client_server_side_apply(fake_api: ResourceApi, mock_client: K8sClient):
    mock_client.request.return_value = fake_pod()
    assert fake_api.server_side_apply(MagicMock(), "name", "namespace", True) == fake_pod()


def test_k8s_client_watch(fake_api: ResourceApi, mock_client: K8sClient):
    resp = MagicMock()
    resp.__iter__.return_value = [
        '{"type": "ADDED", "object": {"apiVersion": "v1", "kind": "A", "data": 1}}',
        '{"type": "DELETED", "object": {"apiVersion": "v1", "kind": "A", "data": 1}}',
        '{"type": "MODIFIED", "object": {"apiVersion": "v1", "kind": "A", "data": 1}}',
        '{"type": "ERROR", "object": {"apiVersion": "v1", "kind": "A", "data": 1}}',
    ]
    fake_api.resource_type = ResourceValue
    mock_client.request.return_value = resp
    for idx, item in enumerate(
        fake_api.watch(
            "namespace",
            "name",
            "label_selector",
            "field_selector",
            "resource_version",
            10,
            None,
        )
    ):
        assert item == pydantic.parse_raw_as(Event, resp.__iter__.return_value[idx])
        assert isinstance(item.object, ResourceValue)
        assert item.object.data == 1



# ruff: noqa: E402
"""All kubernetes lib dependencies collected here."""


__all__ = [
    "ResourceApi",
    "LazyDiscoverer",
    "ApiClient",
    "Configuration",
    "KUBE_CONFIG_DEFAULT_LOCATION",
    "ConfigException",
    "SERVICE_CERT_FILENAME",
    "SERVICE_TOKEN_FILENAME",
    "InClusterConfigLoader",
    "KubeConfigLoader",
    "_get_kube_config_loader",
]

from .kube.discovery import LazyDiscoverer
from .kube.incluster_config import SERVICE_CERT_FILENAME, SERVICE_TOKEN_FILENAME, InClusterConfigLoader
from .kube.kube_config import (
    KUBE_CONFIG_DEFAULT_LOCATION,
    ConfigException,
    KubeConfigLoader,
    _get_kube_config_loader,
)
from .kube.resource_api import ResourceApi as ResourceApi
from .openapi_client.api_client import ApiClient
from .openapi_client.configuration import Configuration



from kubernetes_dynamic.models.resource_value import ResourceValue


def test_resource_field_init():
    item = ResourceValue(key1="val1", key2={"subkey1": 3}, key3=[{"subl1": 4}])
    assert item.key1 == "val1"
    assert isinstance(item.key2, ResourceValue)
    assert item.key2.subkey1 == 3
    assert isinstance(item.key3[0], ResourceValue)
    assert item.key3[0].subl1 == 4


def test_resource_dict_methods():
    item = ResourceValue(key1="val1", key2={"subkey1": 3}, key3=[{"subl1": 4}])
    assert item.to_dict() == dict(key1="val1", key2={"subkey1": 3}, key3=[{"subl1": 4}])
    assert str(item) == item.to_str()
    assert item.get("key1") == "val1"
    assert item.get("key6") is None
    item["key5"] = "newval"
    assert item["key5"] == "newval"



from unittest.mock import MagicMock

from kubernetes_dynamic.config import MutedSecretsSettingsSource


def test_pydantic_mute_no_secrets():
    source = MagicMock(secrets_dir=None)
    settings = MagicMock()
    settings_source = MutedSecretsSettingsSource(source)
    assert settings_source(settings) == {}


def test_pydantic_mute_path(mocker):
    mocker.patch("kubernetes_dynamic.config.Path").return_value.expanduser.return_value.exists.return_value = False
    source = MagicMock(secrets_dir="test/path")
    settings = MagicMock()
    settings_source = MutedSecretsSettingsSource(source)
    assert settings_source(settings) == {}


def test_pydantic_mute_full(mocker):
    mocker.patch("kubernetes_dynamic.config.Path")
    source = MagicMock(secrets_dir="test/path")
    settings = MagicMock()
    settings_source = MutedSecretsSettingsSource(source)
    assert settings_source(settings) == source.return_value



__all__ = [
    "K8sClient",
    "K8sConfig",
    "models",
    "exceptions",
    "cl",
    "CheckResult",
    "ResourceItem",
    "ResourceValue",
    "ResourceApi",
    "Event",
    "EventType",
    "Watch",
    "__version__"
]

from . import exceptions, models
from ._version import __version__
from .client import K8sClient
from .config import K8sConfig
from .events import Event, EventType, Watch
from .kube.resource_api import ResourceApi
from .models.resource_item import CheckResult, ResourceItem
from .models.resource_value import ResourceValue

cl: K8sClient


def __getattr__(name: str):
    if name == "cl":
        global cl
        cl = K8sClient()
        return cl
    raise AttributeError(name)




import http
from enum import Enum
from typing import Any, Dict, Generic, TypeVar

import pydantic
from urllib3 import HTTPResponse

from kubernetes_dynamic.exceptions import ApiException, api_exception
from kubernetes_dynamic.models.resource_value import ResourceValue

HTTP_STATUS_GONE = http.HTTPStatus.GONE


R = TypeVar("R", bound=ResourceValue)


class EventType(Enum):
    """Kubernetes Event type.

    This is related to how Kubernetes events are managed in Kubernetes.
    * ADDED - when a Kubernetes Event is created
    * MODIFIED - when a Kubernetes Event is modified
    * DELETED - when a Kubernetes Event is deleted
    """

    ADDED = "ADDED"
    MODIFIED = "MODIFIED"
    DELETED = "DELETED"
    ERROR = "ERROR"


class Event(ResourceValue, Generic[R]):
    """Kubernetes event object."""

    type: EventType
    object: R
    raw_object: Dict[str, Any]

    @pydantic.root_validator(pre=True)
    def build_extra(cls, values: Dict[str, Any]) -> Dict[str, Any]:  # noqa: B902, N805
        """Automatically add raw_object from object."""
        if "raw_object" not in values:
            values["raw_object"] = values["object"] if isinstance(values["object"], dict) else values["object"].dict()
        return values


class Watch(object):
    def __init__(self, api_client, return_type=None):
        from kubernetes_dynamic.models.resource_item import ResourceItem

        self._return_type = return_type or ResourceItem
        self._stop = False
        self._api_client = api_client
        self.resource_version = None
        self.timeout_seconds = None

    def stop(self):
        self._stop = True

    def stream(
        self,
        func,
        *args,
        resource_version=None,
        watch=True,
        _preload_content=False,
        timeout_seconds=None,
        **kwargs,
    ):
        self._stop = False
        self.resource_version = resource_version or self.resource_version
        self.timeout_seconds = timeout_seconds
        retry_after_410 = False
        while not self._stop:
            resp: HTTPResponse = func(
                *args,
                resource_version=self.resource_version,
                watch=watch,
                _preload_content=_preload_content,
                timeout_seconds=timeout_seconds or 5,
                **kwargs,
            )
            try:
                yield from self._parse_response_iter(resp)
            except ApiException as e:
                if e.status == 410 and not retry_after_410:
                    self.resource_version = func(*args, **kwargs).metadata.resourceVersion
                    retry_after_410 = True
                    continue
                raise api_exception(e) from e
            finally:
                resp.close()
                resp.release_conn()
                if timeout_seconds or self.resource_version is None:
                    self._stop = True

    def _parse_response_iter(self, resp: HTTPResponse):
        for line in resp:
            event = pydantic.parse_raw_as(Event, line)
            if event.type == EventType.ERROR:
                raise ApiException(event.object.code)

            event.object = self._return_type(event.object)
            self.resource_version = event.raw_object.get("metadata", {}).get("resourceVersion")
            yield event

            if self._stop:
                break



from unittest.mock import MagicMock

import pytest

from kubernetes_dynamic.models.secret import V1Secret


def test_secret_init():
    secret = V1Secret(metadata={"name": "my_secret"}, data={"key": "value"})
    assert secret.metadata.name == "my_secret"
    assert secret.data["key"] == "value"


def test_secret_exists(mock_client):
    secret = V1Secret(metadata={"name": "my_secret"})
    assert secret.exists()
    mock_client.get_api.return_value.get.return_value = None
    assert not secret.exists()


@pytest.mark.parametrize(
    "secret, expected",
    [
        (MagicMock(data={"key3": ""}, type="Opaque"), (True, False, ["key1", "key2"])),
        (MagicMock(data={}, type="Opaque"), (True, False, ["key1", "key2", "key3"])),
        (MagicMock(data={}, type="Other"), (True, False, ["key1", "key2", "key3"])),
        (None, (False, False, [])),
        (MagicMock(data={"key3": "", "key2": "", "key1": ""}, type="Opaque"), (True, True, [])),
    ],
)
def test_secret_validate_keys(secret, expected, mock_client):
    mock_client.get_api.return_value.get.return_value = secret
    secret = V1Secret(type="Opaque", metadata={"name": "my_secret"}, required_keys=("key1", "key2", "key3"))
    assert secret.validate_keys() == expected


def test_secret_decode():
    secret = V1Secret(metadata={"name": "my_secret"}, data={"key1": "dGVzdCBzdHJpbmc="})
    assert secret.decode("key1") == "test string"
    assert secret.decode("key_non") == ""


def test_secret_update_empty():
    secret = V1Secret(metadata={"name": "my_secret"})
    secret.set("key1", "test string")
    assert secret.data["key1"] == "dGVzdCBzdHJpbmc="


def test_secret_update_binary():
    secret = V1Secret(metadata={"name": "my_secret"})
    secret.set("key_binary", b"\a4")
    assert secret.data["key_binary"] == "BzQ="
    assert secret.decode("key_binary", "bytes") == b"\a4"
    assert secret.decode("key_binary", "") == b"\a4"


def test_secret_update_same_key():
    secret = V1Secret(metadata={"name": "my_secret"}, data={"key1": "dGVzdCBzdHJpbmc="})
    secret.set("key1", "test string 2")
    assert secret.data["key1"] == "dGVzdCBzdHJpbmcgMg=="


def test_secret_update_other_empty_key():
    secret = V1Secret(metadata={"name": "my_secret"}, data={"key1": "dGVzdCBzdHJpbmc="})
    secret.set("key2", "test string 2")
    assert secret.data["key1"] == "dGVzdCBzdHJpbmc="
    assert secret.data["key2"] == "dGVzdCBzdHJpbmcgMg=="


def test_secret_update_other_key():
    secret = V1Secret(
        metadata={"name": "my_secret"},
        data={"key1": "dGVzdCBzdHJpbmc=", "key2": "dGVzdCBzdHJpbmcgMg=="},
    )
    secret.set("key1", "other test string")
    assert secret.data["key1"] == "b3RoZXIgdGVzdCBzdHJpbmc="
    assert secret.data["key2"] == "dGVzdCBzdHJpbmcgMg=="



from __future__ import annotations

from typing import TYPE_CHECKING, Iterator, Optional

from more_itertools import first
from pydantic import Field
from urllib3 import HTTPResponse

from .common import get_default
from .resource_item import ResourceItem

if TYPE_CHECKING:
    from .all import V1PodSpec, V1PodStatus


class V1Pod(ResourceItem):
    spec: V1PodSpec = Field(default_factory=lambda: get_default("V1PodSpec"))
    status: V1PodStatus = Field(default_factory=lambda: get_default("V1PodStatus"))

    @classmethod
    def get_restarts(cls):
        """Get pod restarts."""
        pod_restarts: dict[str, dict] = {}
        for pod in cls.default_client().pods.get():
            pod_restarts[pod.metadata.name] = {}
            for container in pod.status.containerStatuses:
                if container.restartCount > 0:
                    pod_restarts[pod.metadata.name][container.name] = {
                        "restart_count": container.restartCount,
                        "last_restart_reason": container.lastState.terminated.reason,
                        "last_restart_finished_at": container.lastState.terminated.finishedAt,
                    }
                else:
                    pod_restarts[pod.metadata.name][container.name] = {
                        "restart_count": container.restartCount,
                        "last_restart_reason": "N/A",
                        "last_restart_finished_at": "N/A",
                    }
        return pod_restarts

    def exec(
        self,
        command: str | list[str],
        container: str = "",
        stdin: bool = True,
        stdout: bool = True,
        stderr: bool = True,
        tty: bool = True,
    ) -> str:
        """Run command on pod."""
        response = self._client.stream(
            self._api.exec.get,  # type: ignore
            self.metadata.name,
            self.metadata.namespace,
            container=container,
            command=command,
            stdin=stdin,
            stdout=stdout,
            stderr=stderr,
            tty=tty,
        )
        return response

    def logs(
        self,
        container: Optional[str] = None,
        follow: Optional[bool] = None,
        insecure: Optional[bool] = None,
        limit: Optional[int] = None,
        pretty: Optional[bool] = None,
        previous: Optional[bool] = None,
        since_seconds: Optional[int] = None,
        tail_lines: Optional[int] = None,
        timestamps: Optional[bool] = None,
    ) -> Iterator[str]:
        """Get pod logs."""
        response: HTTPResponse = self._api.log.get(
            self.metadata.name,
            self.metadata.namespace,
            container=container,
            follow=follow,
            insecureSkipTLSVerifyBackend=insecure,
            limitBytes=limit,
            pretty=pretty,
            previous=previous,
            sinceSeconds=since_seconds,
            tailLines=tail_lines,
            timestamps=timestamps,
            serialize=False,
            stream=True,
        )  # type: ignore
        for data in response.stream():
            for line in data.splitlines():
                yield line.decode().strip()

    def disk_usage(self, container: str = "") -> dict[str, dict[str, int]]:
        """Get disc usage on a pod's container."""
        data = {}
        output = self.exec(["df", "--output=used,size,avail,pcent,target"], container)
        for line in output.splitlines()[1:]:
            used, size, avail, pcent, target = line.split()
            data[target] = {
                "used": int(used),
                "size": int(size),
                "avail": int(avail),
                "pcent": int(pcent.strip("%")),
            }
        return data

    def get_controller_type(self) -> str:
        """Get pod controller type."""
        controller = self.metadata.ownerReferences[0].kind
        controller = "Deployment" if controller == "ReplicaSet" else controller
        return controller

    def get_env(self) -> dict[str, str]:
        """Get environment variables from a pod."""
        env = self.exec("env")
        return {item.split("=", 1)[0]: item.split("=", 1)[1] for item in env.splitlines()}

    def get_port(self, port: str | int) -> Optional[int]:
        """Get the first matching port number with name."""
        if isinstance(port, int):
            return first(
                (p.containerPort for c in self.spec.containers for p in c.ports if p.containerPort == port), None
            )
        return first(
            (p.containerPort for c in self.spec.containers for p in c.ports if p.name == port or not port), None
        )



from __future__ import annotations

from typing import Any, Dict

import pydantic
import yaml


class ResourceValue(pydantic.BaseModel):
    """Every kubernetes field with dictionary like format."""

    class Config:
        """Pydantic config class."""

        extra = "allow"
        arbitrary_types_allowed = True
        use_enum_values = True

    def __init__(
        self,
        definition: dict[str, Any] | ResourceValue | None = None,
        **kwargs,
    ):
        final_def = ResourceValue.merge_definition_with_kwargs(definition, **kwargs)
        super().__init__(**final_def)

    @classmethod
    def merge_definition_with_kwargs(
        cls,
        definition: dict | ResourceValue | None = None,
        **kwargs,
    ) -> dict[str, Any]:
        final_def: dict = {}
        definition = definition or {}
        if isinstance(definition, dict):
            final_def = definition
        elif isinstance(definition, ResourceValue):
            final_def = definition.dict()  # type: ignore
        final_def.update(kwargs)
        return final_def

    @pydantic.root_validator(pre=True)
    def build_extra(cls, values: Dict[str, Any]) -> Dict[str, Any]:  # noqa: B902, N805
        """Automatically convert extra items to ResourceValues."""
        field_names = {field.alias for field in cls.__fields__.values()}

        def convert(item: Any):
            if isinstance(item, dict):
                return ResourceValue(**item)
            if isinstance(item, list):
                return [convert(it) for it in item]
            return item

        for name, value in values.items():
            if name in field_names:
                continue
            values[name] = convert(value)
        return values

    def _update_attrs(self, data: dict | ResourceValue):
        data = self.validate(data)
        for key in data.__fields__:
            setattr(self, key, getattr(data, key))
        return self

    def __str__(self):
        return "{}:\n  {}".format(self.__class__.__name__, "  ".join(yaml.safe_dump(self.dict()).splitlines(True)))

    def __getattr__(self, name: str) -> Any:
        return None

    def __getitem__(self, name: str):
        return self.__dict__[name]

    def __setitem__(self, name: str, value: Any):
        setattr(self, name, value)

    def keys(self):
        return self.__dict__.keys()

    def to_dict(self):
        """Convert to dict."""
        return self.dict()

    def to_str(self):
        """Get a yaml string representation."""
        return str(self)

    def get(self, name: str, default: Any = None):
        """Same as dict.get."""
        return self.__dict__.get(name, default)

    def __contains__(self, m):
        return m in self.__dict__

    def pop(self, key, default):
        """Same as dict.pop."""
        return self.__dict__.pop(key, default)



# Copyright 2018 The Kubernetes Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import select
import socket
import ssl
import struct
import threading
import time
from base64 import urlsafe_b64decode
from io import StringIO
from urllib.parse import urlencode, urlparse, urlunparse

import certifi
import yaml
from websocket import ABNF, WebSocket, enableTrace

from kubernetes_dynamic.openapi_client.exceptions import ApiException, ApiValueError

STDIN_CHANNEL = 0
STDOUT_CHANNEL = 1
STDERR_CHANNEL = 2
ERROR_CHANNEL = 3
RESIZE_CHANNEL = 4


class _IgnoredIO:
    def write(self, _x):
        pass

    def getvalue(self):
        raise TypeError(
            "Tried to read_all() from a WSClient configured to not capture. Did you mean `capture_all=True`?"
        )


class WSClient:
    def __init__(self, configuration, url, headers, capture_all):
        """A websocket client with support for channels.

            Exec command uses different channels for different streams. for
        example, 0 is stdin, 1 is stdout and 2 is stderr. Some other API calls
        like port forwarding can forward different pods' streams to different
        channels.
        """
        self._connected = False
        self._channels = {}
        if capture_all:
            self._all = StringIO()
        else:
            self._all = _IgnoredIO()
        self.sock = create_websocket(configuration, url, headers)
        self._connected = True
        self._returncode = None

    def peek_channel(self, channel, timeout=0):
        """Peek a channel and return part of the input,
        empty string otherwise."""
        self.update(timeout=timeout)
        if channel in self._channels:
            return self._channels[channel]
        return ""

    def read_channel(self, channel, timeout=0):
        """Read data from a channel."""
        if channel not in self._channels:
            ret = self.peek_channel(channel, timeout)
        else:
            ret = self._channels[channel]
        if channel in self._channels:
            del self._channels[channel]
        return ret

    def readline_channel(self, channel, timeout=None):
        """Read a line from a channel."""
        if timeout is None:
            timeout = float("inf")
        start = time.time()
        while self.is_open() and time.time() - start < timeout:
            if channel in self._channels:
                data = self._channels[channel]
                if "\n" in data:
                    index = data.find("\n")
                    ret = data[:index]
                    data = data[index + 1 :]
                    if data:
                        self._channels[channel] = data
                    else:
                        del self._channels[channel]
                    return ret
            self.update(timeout=(timeout - time.time() + start))

    def write_channel(self, channel, data):
        """Write data to a channel."""
        # check if we're writing binary data or not
        binary = isinstance(data, bytes)
        opcode = ABNF.OPCODE_BINARY if binary else ABNF.OPCODE_TEXT

        channel_prefix = chr(channel)
        if binary:
            channel_prefix = bytes(channel_prefix, "ascii")

        payload = channel_prefix + data
        self.sock.send(payload, opcode=opcode)

    def peek_stdout(self, timeout=0):
        """Same as peek_channel with channel=1."""
        return self.peek_channel(STDOUT_CHANNEL, timeout=timeout)

    def read_stdout(self, timeout=None):
        """Same as read_channel with channel=1."""
        return self.read_channel(STDOUT_CHANNEL, timeout=timeout)

    def readline_stdout(self, timeout=None):
        """Same as readline_channel with channel=1."""
        return self.readline_channel(STDOUT_CHANNEL, timeout=timeout)

    def peek_stderr(self, timeout=0):
        """Same as peek_channel with channel=2."""
        return self.peek_channel(STDERR_CHANNEL, timeout=timeout)

    def read_stderr(self, timeout=None):
        """Same as read_channel with channel=2."""
        return self.read_channel(STDERR_CHANNEL, timeout=timeout)

    def readline_stderr(self, timeout=None):
        """Same as readline_channel with channel=2."""
        return self.readline_channel(STDERR_CHANNEL, timeout=timeout)

    def read_all(self):
        """Return buffered data received on stdout and stderr channels.
        This is useful for non-interactive call where a set of command passed
        to the API call and their result is needed after the call is concluded.
        Should be called after run_forever() or update()

        TODO: Maybe we can process this and return a more meaningful map with
        channels mapped for each input.
        """
        out = self._all.getvalue()
        self._all = self._all.__class__()
        self._channels = {}
        return out

    def is_open(self):
        """True if the connection is still alive."""
        return self._connected

    def write_stdin(self, data):
        """The same as write_channel with channel=0."""
        self.write_channel(STDIN_CHANNEL, data)

    def update(self, timeout=0):
        """Update channel buffers with at most one complete frame of input."""
        if not self.is_open():
            return
        if not self.sock.connected:
            self._connected = False
            return

        # The options here are:
        # select.select() - this will work on most OS, however, it has a
        #                   limitation of only able to read fd numbers up to 1024.
        #                   i.e. does not scale well. This was the original
        #                   implementation.
        # select.poll()   - this will work on most unix based OS, but not as
        #                   efficient as epoll. Will work for fd numbers above 1024.
        # select.epoll()  - newest and most efficient way of polling.
        #                   However, only works on linux.
        if hasattr(select, "poll"):
            poll = select.poll()
            poll.register(self.sock.sock, select.POLLIN)
            if timeout is not None:
                timeout *= 1_000  # poll method uses milliseconds as the time unit
            r = poll.poll(timeout)
            poll.unregister(self.sock.sock)
        else:
            r, _, _ = select.select((self.sock.sock,), (), (), timeout)

        if r:
            op_code, frame = self.sock.recv_data_frame(True)
            if op_code == ABNF.OPCODE_CLOSE:
                self._connected = False
                return
            elif op_code == ABNF.OPCODE_BINARY or op_code == ABNF.OPCODE_TEXT:
                data = frame.data
                data = data.decode("utf-8", "replace")
                if len(data) > 1:
                    channel = ord(data[0])
                    data = data[1:]
                    if data:
                        if channel in [STDOUT_CHANNEL, STDERR_CHANNEL]:
                            # keeping all messages in the order they received
                            # for non-blocking call.
                            self._all.write(data)
                        if channel not in self._channels:
                            self._channels[channel] = data
                        else:
                            self._channels[channel] += data

    def run_forever(self, timeout=None):
        """Wait till connection is closed or timeout reached. Buffer any input
        received during this time."""
        if timeout:
            start = time.time()
            while self.is_open() and time.time() - start < timeout:
                self.update(timeout=(timeout - time.time() + start))
        else:
            while self.is_open():
                self.update(timeout=None)

    @property
    def returncode(self):
        """
        The return code, A None value indicates that the process hasn't
        terminated yet.
        """
        if self.is_open():
            return None
        else:
            if self._returncode is None:
                err = self.read_channel(ERROR_CHANNEL)
                err = yaml.safe_load(err)
                if err["status"] == "Success":
                    self._returncode = 0
                else:
                    self._returncode = int(err["details"]["causes"][0]["message"])
            return self._returncode

    def close(self, **kwargs):
        """
        close websocket connection.
        """
        self._connected = False
        if self.sock:
            self.sock.close(**kwargs)


WSResponse = collections.namedtuple("WSResponse", ["data"])


class PortForward:
    def __init__(self, websocket, ports):
        """A websocket client with support for port forwarding.

        Port Forward command sends on 2 channels per port, a read/write
        data channel and a read only error channel. Both channels are sent an
        initial frame contaning the port number that channel is associated with.
        """

        self.websocket = websocket
        self.local_ports = {}
        for ix, port_number in enumerate(ports):
            self.local_ports[port_number] = self._Port(ix, port_number)
        # There is a thread run per PortForward instance which performs the translation between the
        # raw socket data sent by the python application and the websocket protocol. This thread
        # terminates after either side has closed all ports, and after flushing all pending data.
        proxy = threading.Thread(
            name="Kubernetes port forward proxy: %s" % ", ".join([str(port) for port in ports]), target=self._proxy
        )
        proxy.daemon = True
        proxy.start()

    @property
    def connected(self):
        return self.websocket.connected

    def socket(self, port_number):
        if port_number not in self.local_ports:
            raise ValueError("Invalid port number")
        return self.local_ports[port_number].socket

    def error(self, port_number):
        if port_number not in self.local_ports:
            raise ValueError("Invalid port number")
        return self.local_ports[port_number].error

    def close(self):
        for port in self.local_ports.values():
            port.socket.close()

    class _Port:
        def __init__(self, ix, port_number):
            # The remote port number
            self.port_number = port_number
            # The websocket channel byte number for this port
            self.channel = struct.Struct(">B").pack(ix * 2)
            # A socket pair is created to provide a means of translating the data flow
            # between the python application and the kubernetes websocket. The self.python
            # half of the socket pair is used by the _proxy method to receive and send data
            # to the running python application.
            s, self.python = socket.socketpair()
            # The self.socket half of the pair is used by the python application to send
            # and receive data to the eventual pod port. It is wrapped in the _Socket class
            # because a socket pair is an AF_UNIX socket, not a AF_INET socket. This allows
            # intercepting setting AF_INET socket options that would error against an AF_UNIX
            # socket.
            self.socket = self._Socket(s)
            # Data accumulated from the websocket to be sent to the python application.
            self.data = b""
            # All data sent from kubernetes on the port error channel.
            self.error = None

        class _Socket:
            def __init__(self, socket):
                self._socket = socket

            def __getattr__(self, name):
                return getattr(self._socket, name)

            def setsockopt(self, level, optname, value):
                # The following socket option is not valid with a socket created from socketpair,
                # and is set by the http.client.HTTPConnection.connect method.
                if level == socket.IPPROTO_TCP and optname == socket.TCP_NODELAY:
                    return
                self._socket.setsockopt(level, optname, value)

    # Proxy all socket data between the python code and the kubernetes websocket.
    def _proxy(self):
        channel_ports = []
        channel_initialized = []
        local_ports = {}
        for port in self.local_ports.values():
            # Setup the data channel for this port number
            channel_ports.append(port)
            channel_initialized.append(False)
            # Setup the error channel for this port number
            channel_ports.append(port)
            channel_initialized.append(False)
            port.python.setblocking(True)
            local_ports[port.python] = port
        # The data to send on the websocket socket
        kubernetes_data = b""
        while True:
            rlist = []  # List of sockets to read from
            wlist = []  # List of sockets to write to
            if self.websocket.connected:
                rlist.append(self.websocket)
                if kubernetes_data:
                    wlist.append(self.websocket)
            local_all_closed = True
            for port in self.local_ports.values():
                if port.python.fileno() != -1:
                    if self.websocket.connected:
                        rlist.append(port.python)
                        if port.data:
                            wlist.append(port.python)
                        local_all_closed = False
                    else:
                        if port.data:
                            wlist.append(port.python)
                            local_all_closed = False
                        else:
                            port.python.close()
            if local_all_closed and not (self.websocket.connected and kubernetes_data):
                self.websocket.close()
                return
            r, w, _ = select.select(rlist, wlist, [])
            for sock in r:
                if sock == self.websocket:
                    pending = True
                    while pending:
                        opcode, frame = self.websocket.recv_data_frame(True)
                        if opcode == ABNF.OPCODE_BINARY:
                            if not frame.data:
                                raise RuntimeError("Unexpected frame data size")
                            channel = int(frame.data)
                            if channel >= len(channel_ports):
                                raise RuntimeError("Unexpected channel number: %s" % channel)
                            port = channel_ports[channel]
                            if channel_initialized[channel]:
                                if channel % 2:
                                    if port.error is None:
                                        port.error = ""
                                    port.error += frame.data[1:].decode()
                                    port.python.close()
                                else:
                                    port.data += frame.data[1:]
                            else:
                                if len(frame.data) != 3:
                                    raise RuntimeError("Unexpected initial channel frame data size")
                                port_number = int(frame.data[1:2]) + (int(frame.data[2:3]) * 256)
                                if port_number != port.port_number:
                                    raise RuntimeError(
                                        "Unexpected port number in initial channel frame: %s" % port_number
                                    )
                                channel_initialized[channel] = True
                        elif opcode not in (ABNF.OPCODE_PING, ABNF.OPCODE_PONG, ABNF.OPCODE_CLOSE):
                            raise RuntimeError("Unexpected websocket opcode: %s" % opcode)
                        if not (isinstance(self.websocket.sock, ssl.SSLSocket) and self.websocket.sock.pending()):
                            pending = False
                else:
                    port = local_ports[sock]
                    if port.python.fileno() != -1:
                        data = port.python.recv(1024 * 1024)
                        if data:
                            kubernetes_data += ABNF.create_frame(
                                port.channel + data,
                                ABNF.OPCODE_BINARY,
                            ).format()
                        else:
                            port.python.close()
            for sock in w:
                if sock == self.websocket:
                    sent = self.websocket.sock.send(kubernetes_data)
                    kubernetes_data = kubernetes_data[sent:]
                else:
                    port = local_ports[sock]
                    if port.python.fileno() != -1:
                        sent = port.python.send(port.data)
                        port.data = port.data[sent:]


def get_websocket_url(url, query_params=None):
    parsed_url = urlparse(url)
    parts = list(parsed_url)
    if parsed_url.scheme == "http":
        parts[0] = "ws"
    elif parsed_url.scheme == "https":
        parts[0] = "wss"
    if query_params:
        query = []
        for key, value in query_params:
            if key == "command" and isinstance(value, list):
                for command in value:
                    query.append((key, command))
            else:
                query.append((key, value))
        if query:
            parts[4] = urlencode(query)
    return urlunparse(parts)


def create_websocket(configuration, url, headers=None):
    enableTrace(False)

    # We just need to pass the Authorization, ignore all the other
    # http headers we get from the generated code
    header = []
    if headers and "authorization" in headers:
        header.append("authorization: %s" % headers["authorization"])
    if headers and "sec-websocket-protocol" in headers:
        header.append("sec-websocket-protocol: %s" % headers["sec-websocket-protocol"])
    else:
        header.append("sec-websocket-protocol: v4.channel.k8s.io")

    if url.startswith("wss://") and configuration.verify_ssl:
        ssl_opts = {
            "cert_reqs": ssl.CERT_REQUIRED,
            "ca_certs": configuration.ssl_ca_cert or certifi.where(),
        }
        if configuration.assert_hostname is not None:
            ssl_opts["check_hostname"] = configuration.assert_hostname
    else:
        ssl_opts = {"cert_reqs": ssl.CERT_NONE}

    if configuration.cert_file:
        ssl_opts["certfile"] = configuration.cert_file
    if configuration.key_file:
        ssl_opts["keyfile"] = configuration.key_file

    websocket = WebSocket(sslopt=ssl_opts, skip_utf8_validation=False)
    connect_opt = {"header": header}

    if configuration.proxy or configuration.proxy_headers:
        connect_opt = websocket_proxycare(connect_opt, configuration, url, headers)

    websocket.connect(url, **connect_opt)
    return websocket


def websocket_proxycare(connect_opt, configuration, url, headers):
    """An internal function to be called in api-client when a websocket
    create is requested.
    """
    if configuration.no_proxy:
        connect_opt.update({"http_no_proxy": configuration.no_proxy.split(",")})

    if configuration.proxy:
        proxy_url = urlparse(configuration.proxy)
        connect_opt.update({"http_proxy_host": proxy_url.hostname, "http_proxy_port": proxy_url.port})
    if configuration.proxy_headers:
        for key, value in configuration.proxy_headers.items():
            if key == "proxy-authorization" and value.startswith("Basic"):
                b64value = value.split()[1]
                auth = urlsafe_b64decode(b64value).decode().split(":")
                connect_opt.update({"http_proxy_auth": (auth[0], auth[1])})
    return connect_opt


def websocket_call(configuration, _method, url, **kwargs):
    """An internal function to be called in api-client when a websocket
    connection is required. method, url, and kwargs are the parameters of
    apiClient.request method."""

    url = get_websocket_url(url, kwargs.get("query_params"))
    headers = kwargs.get("headers")
    _request_timeout = kwargs.get("_request_timeout", 60)
    _preload_content = kwargs.get("_preload_content", True)
    capture_all = kwargs.get("capture_all", True)

    try:
        client = WSClient(configuration, url, headers, capture_all)
        if not _preload_content:
            return client
        client.run_forever(timeout=_request_timeout)
        return WSResponse("%s" % "".join(client.read_all()))
    except (Exception, KeyboardInterrupt, SystemExit) as e:
        raise ApiException(status=0, reason=str(e))


def portforward_call(configuration, _method, url, **kwargs):
    """An internal function to be called in api-client when a websocket
    connection is required for port forwarding. args and kwargs are the
    parameters of apiClient.request method."""

    query_params = kwargs.get("query_params")

    ports = []
    for param, value in query_params:
        if param == "ports":
            for port in value.split(","):
                try:
                    port_number = int(port)
                except ValueError:
                    raise ApiValueError("Invalid port number: %s" % port)
                if not (0 < port_number < 65536):
                    raise ApiValueError("Port number must be between 0 and 65536: %s" % port)
                if port_number in ports:
                    raise ApiValueError("Duplicate port numbers: %s" % port)
                ports.append(port_number)
    if not ports:
        raise ApiValueError("Missing required parameter `ports`")

    url = get_websocket_url(url, query_params)
    headers = kwargs.get("headers")

    try:
        websocket = create_websocket(configuration, url, headers)
        return PortForward(websocket, ports)
    except (Exception, KeyboardInterrupt, SystemExit) as e:
        raise ApiException(status=0, reason=str(e))



from __future__ import annotations

import json
from pathlib import Path
from typing import Any, Callable, List, Optional, Type, TypeVar, overload

import pydantic
import yaml

import kubernetes_dynamic.kube.ws_client as ws_client
import kubernetes_dynamic.models as models
from kubernetes_dynamic.formatters import to_lower_camel

from . import _kubernetes
from .config import K8sConfig
from .exceptions import (
    ConfigException,
    InvalidParameter,
    ResourceNotUniqueError,
)
from .kube.resource_api import MISSING, ResourceApi, _Missing
from .models.common import ItemList, get_type
from .models.resource_item import ResourceItem

T = TypeVar("T", bound=ResourceItem)


def serialize_object(
    data, serializer: Optional[Type[ResourceItem] | Callable] = None
) -> ResourceItem | ItemList[ResourceItem]:
    if serializer and not isinstance(serializer, Type):
        try:
            return serializer(data)
        except TypeError:
            return serializer(data, data)
    kind = data.get("kind", "")
    is_list = False
    if kind.endswith("List") and "items" in data:
        kind = kind[:-4]
        is_list = True

    api_version = data.get("apiVersion", "v1")

    obj_type = serializer or get_type(kind, api_version, ResourceItem)

    if not is_list:
        return pydantic.parse_obj_as(obj_type, data)

    for item in data["items"]:
        item.setdefault("apiVersion", api_version)
        item.setdefault("kind", kind)
    items = pydantic.parse_obj_as(List[obj_type], data["items"])
    return ItemList(items, metadata=data["metadata"])


def meta_request(func):
    """Handles parsing response structure and translating API Exceptions"""

    def inner(self, *args, **kwargs):
        serialize = kwargs.pop("serialize", True)
        serializer = kwargs.pop("serializer", None)
        response = func(self, *args, **kwargs)
        if not response:
            return None
        if not serialize:
            return response
        try:
            data = json.loads(response.data)
            return serialize_object(data, serializer)
        except json.JSONDecodeError:
            return response.data

    return inner


class K8sClient(object):
    _loaded: bool = False
    in_cluster: bool = False

    componentstatuses: ResourceApi[models.V1ComponentStatus]
    configmaps: ResourceApi[models.V1ConfigMap]
    endpoints: ResourceApi[models.V1Endpoints]
    events: ResourceApi[models.CoreV1Event]
    limitranges: ResourceApi[models.V1LimitRange]
    namespaces: ResourceApi[models.V1Namespace]
    nodes: ResourceApi[models.V1Node]
    persistentvolumeclaims: ResourceApi[models.V1PersistentVolumeClaim]
    persistentvolumes: ResourceApi[models.V1PersistentVolume]
    pods: ResourceApi[models.V1Pod]
    podtemplates: ResourceApi[models.V1PodTemplate]
    replicationcontrollers: ResourceApi[models.V1ReplicationController]
    resourcequotas: ResourceApi[models.V1ResourceQuota]
    secrets: ResourceApi[models.V1Secret]
    serviceaccounts: ResourceApi[models.V1ServiceAccount]
    services: ResourceApi[models.V1Service]
    mutatingwebhookconfigurations: ResourceApi[models.V1MutatingWebhookConfiguration]
    validatingwebhookconfigurations: ResourceApi[models.V1ValidatingWebhookConfiguration]
    customresourcedefinitions: ResourceApi[models.V1CustomResourceDefinition]
    apiservices: ResourceApi[models.V1APIService]
    controllerrevisions: ResourceApi[models.V1ControllerRevision]
    daemonsets: ResourceApi[models.V1DaemonSet]
    deployments: ResourceApi[models.V1Deployment]
    replicasets: ResourceApi[models.V1ReplicaSet]
    statefulsets: ResourceApi[models.V1StatefulSet]
    horizontalpodautoscalers: ResourceApi[models.V1HorizontalPodAutoscaler]
    cronjobs: ResourceApi[models.V1CronJob]
    jobs: ResourceApi[models.V1Job]
    certificatesigningrequests: ResourceApi[models.V1CertificateSigningRequest]
    leases: ResourceApi[models.V1Lease]
    endpointslices: ResourceApi[models.V1EndpointSlice]
    flowschemas: ResourceApi[models.V1beta2FlowSchema]
    prioritylevelconfigurations: ResourceApi[models.V1beta2PriorityLevelConfiguration]
    ingressclasses: ResourceApi[models.V1IngressClass]
    ingresses: ResourceApi[models.V1Ingress]
    networkpolicies: ResourceApi[models.V1NetworkPolicy]
    runtimeclasses: ResourceApi[models.V1RuntimeClass]
    poddisruptionbudgets: ResourceApi[models.V1PodDisruptionBudget]
    clusterrolebindings: ResourceApi[models.V1ClusterRoleBinding]
    clusterroles: ResourceApi[models.V1ClusterRole]
    rolebindings: ResourceApi[models.V1RoleBinding]
    roles: ResourceApi[models.V1Role]
    priorityclasses: ResourceApi[models.V1PriorityClass]
    csidrivers: ResourceApi[models.V1CSIDriver]
    csinodes: ResourceApi[models.V1CSINode]
    csistoragecapacities: ResourceApi[models.V1CSIStorageCapacity]
    storageclasses: ResourceApi[models.V1StorageClass]
    volumeattachments: ResourceApi[models.V1VolumeAttachment]

    def __init__(
        self,
        api_client: Optional[_kubernetes.ApiClient] = None,
        *,
        config_file: Optional[str] = None,
        config_dict: Optional[dict] = None,
        context: Optional[str] = None,
        cache_file=None,
        discoverer=None,
    ):
        discoverer = discoverer or _kubernetes.LazyDiscoverer
        self.config = self.get_config(config_file, config_dict=config_dict, context=context)
        self.client = api_client or _kubernetes.ApiClient(configuration=self.config.configuration)
        self.configuration = self.client.configuration
        self.__discoverer = discoverer(self, cache_file)

    @property
    def resources(self):
        return self.__discoverer

    @property
    def version(self):
        return self.__discoverer.version

    @staticmethod
    def get_config_file() -> str:
        """Get kube config file."""
        return str(Path(_kubernetes.KUBE_CONFIG_DEFAULT_LOCATION).expanduser().resolve())

    @staticmethod
    def get_kube_config_loader(
        config_file: Optional[str] = None, config_dict: Optional[dict] = None, context: Optional[str] = None
    ) -> _kubernetes.InClusterConfigLoader | _kubernetes.KubeConfigLoader:
        if not config_dict and not config_file:
            config_file = K8sClient.get_config_file()
        try:
            in_cluster_loader = _kubernetes.InClusterConfigLoader(
                token_filename=_kubernetes.SERVICE_TOKEN_FILENAME,
                cert_filename=_kubernetes.SERVICE_CERT_FILENAME,
                try_refresh_token=True,
            )
            in_cluster_loader.load_and_set()
            K8sClient.in_cluster = True
            return in_cluster_loader
        except ConfigException:
            K8sClient.in_cluster = False
            return _kubernetes._get_kube_config_loader(config_file, config_dict, active_context=context)

    @staticmethod
    def find_context(name: Optional[str], loader: _kubernetes.KubeConfigLoader) -> Optional[str]:
        """Find a context by name, or get the default."""
        if not name:
            return None
        contexts = loader.list_contexts()
        for context in contexts:
            if not isinstance(context, dict):
                continue
            if "context" not in context:
                continue
            if context.get("name") == name or context["context"].get("cluster") == name:
                return context.get("name")
        raise RuntimeError(f"No context name='{name}' found!")

    def get_config(
        self,
        config_file: Optional[str] = None,
        *,
        config_dict: Optional[dict] = None,
        context: Optional[str] = None,
    ) -> K8sConfig:
        """Get kubernetes config.

        Args:
            config_file: File to load.
            config_dict: Load config from dict instead of using a file.
            context: Set kube context.
        """
        loader = K8sClient.get_kube_config_loader(config_file, config_dict, context=context)
        if isinstance(loader, _kubernetes.InClusterConfigLoader):
            return K8sConfig(configuration=_kubernetes.Configuration.get_default_copy())
        config_file = config_file or K8sClient.get_config_file()
        context = K8sClient.find_context(context, loader)
        loader.set_active_context(context)
        configuration = type.__call__(_kubernetes.Configuration)
        loader.load_and_set(configuration)
        return K8sConfig(
            configuration=configuration,
            context=str(loader.current_context["context"].get("cluster")),
            namespace=str(loader.current_context["context"].get("namespace", "default")),
        )

    @overload
    def get_api(
        self,
        name: Optional[str] = None,
        object_type: None = None,
        api_version: Optional[str] = None,
        kind: Optional[str] = None,
        **filter_dict,
    ) -> ResourceApi[ResourceItem]:
        ...

    @overload
    def get_api(
        self,
        name: Optional[str] = None,
        object_type: Type[T] = None,
        api_version: Optional[str] = None,
        kind: Optional[str] = None,
        **filter_dict,
    ) -> ResourceApi[T]:
        ...

    def get_api(
        self,
        name: Optional[str] = None,
        object_type: Optional[Type[T]] = ResourceItem,
        api_version: Optional[str] = None,
        kind: Optional[str] = None,
        **filter_dict,
    ) -> ResourceApi[T]:
        if api_version:
            filter_dict["api_version"] = api_version
        if kind:
            filter_dict["kind"] = kind
        if name:
            filter_dict["name"] = name
        try:
            api = self.resources.get(**filter_dict)
        except ResourceNotUniqueError:
            api = [
                r
                for r in self.resources.search(**filter_dict)
                if r.preferred and isinstance(r, _kubernetes.ResourceApi)
            ][0]
        api.resource_type = object_type or get_type(str(api.kind), str(api.api_version), ResourceItem)  # type: ignore
        return api  # type: ignore

    def __getattr__(self, name: str) -> ResourceApi[ResourceItem]:
        if name.startswith("_"):
            raise AttributeError(name)
        return self.get_api(name)

    @property
    def events_events(self) -> ResourceApi[models.EventsV1Event]:
        return self.get_api("events", object_type=models.EventsV1Event, api_version="events.k8s.io/v1")

    def websocket(self, func: Callable, name=None, namespace=MISSING, *args, **kwargs) -> ws_client.WSClient:
        prev_request = self.client.request
        try:

            def _websocket(*args, **kwargs):  # pragma: no cover
                try:
                    client = ws_client.websocket_call(self.configuration, *args, **kwargs)
                except Exception as e:
                    raise e
                return client

            self.client.request = _websocket
            return func(
                *args,
                name=name,
                namespace=namespace,
                query_params=[(k, v) for k, v in kwargs.items() if v is not None],
                _preload_content=True,
                serialize=False,
            )
        finally:
            self.client.request = prev_request

    def stream(self, func: Callable, name=None, namespace=MISSING, *args, **kwargs) -> str:
        client = self.websocket(func, name, namespace, *args, **kwargs)
        client.run_forever(timeout=kwargs.get("_request_timeout", 0))  # type: ignore
        return ws_client.WSResponse("%s" % "".join(client.read_all())).data

    @meta_request
    def request(self, method: str, path: str, body=None, **params) -> Any:
        if not path.startswith("/"):
            path = "/" + path

        path_params = params.pop("path_params", {})
        query_params = params.pop("query_params", [])

        header_params = params.pop("header_params", {})
        form_params = []
        local_var_files = {}

        # Checking Accept header.
        new_header_params = dict((key.lower(), value) for key, value in header_params.items())
        if "accept" not in new_header_params:
            header_params["Accept"] = "application/json"

        # HTTP header `Content-Type`
        header_params["Content-Type"] = params.pop("content_type", "application/json")
        async_req = params.pop("async_req", False)
        _return_http_data_only = params.pop("_return_http_data_only", True)
        _request_timeout = params.pop("_request_timeout", None)

        # Authentication setting
        auth_settings = ["BearerToken"]

        for key, value in params.items():
            if value is not None:
                query_params.append((to_lower_camel(key.lstrip("_")), value))

        api_response = self.client.call_api(
            path,
            method.upper(),
            path_params,
            query_params,
            header_params,
            body=body,
            post_params=form_params,
            async_req=async_req,
            files=local_var_files,
            auth_settings=auth_settings,
            _preload_content=False,
            _return_http_data_only=_return_http_data_only,
            _request_timeout=_request_timeout,
        )
        return api_response.get() if async_req else api_response  # type: ignore

    def apply(
        self,
        *,
        namespace: Optional[str | _Missing] = MISSING,
        file_path: Optional[str | Path] = None,
        data: Optional[dict | list | ResourceItem] = None,
    ) -> list[ResourceItem]:
        """Apply a kubernetes resource."""
        if file_path is not None and data is not None:
            raise InvalidParameter("`file_path` and `data` is mutually exclusive.")
        if not data:
            if not file_path:
                raise InvalidParameter("`file_path` or `data` must be provided.")
            if not Path(file_path).exists():
                raise FileNotFoundError(file_path)
            with open(file_path) as fp:
                data = list(yaml.full_load_all(fp))
        if not isinstance(data, list):
            data = [data]

        items = []
        for item in data:
            resource = self.get_api(kind=item["kind"], api_version=item["apiVersion"].split("/")[-1])
            items.append(resource.apply(item, namespace))
        return items



from __future__ import annotations

from collections import UserList
from datetime import datetime
from typing import Any, Dict, Generic, Iterator, List, Optional, TypeVar

import pydantic
from pydantic import Field

from .resource_value import ResourceValue

R = TypeVar("R", bound=ResourceValue)


class V1ManagedFieldsEntry(ResourceValue):
    apiVersion: str = "v1"
    fieldsType: Optional[str] = None
    fieldsV1: Optional[object] = None
    manager: Optional[str] = None
    operation: Optional[str] = None
    subresource: Optional[str] = None
    time: Optional[datetime] = None


class V1OwnerReference(ResourceValue):
    apiVersion: str = "v1"
    kind: str = "OwnerReference"
    metadata: V1ObjectMeta = Field(default_factory=lambda: V1ObjectMeta())
    blockOwnerDeletion: Optional[bool] = None
    controller: Optional[bool] = None
    name: str
    uid: Optional[str] = None


class V1ObjectMeta(ResourceValue):
    annotations: Dict[str, str] = Field(default_factory=dict)
    creationTimestamp: Optional[datetime] = None
    deletionGracePeriodSeconds: Optional[int] = None
    deletionTimestamp: Optional[datetime] = None
    finalizers: List[str] = Field(default_factory=list)
    generateName: Optional[str] = None
    generation: Optional[int] = None
    labels: Dict[str, str] = Field(default_factory=dict)
    managedFields: List[V1ManagedFieldsEntry] = Field(default_factory=list)
    name: str = ""
    namespace: Optional[str] = None
    ownerReferences: List[V1OwnerReference] = Field(default_factory=list)
    resourceVersion: Optional[str] = None
    selfLink: Optional[str] = None
    uid: Optional[str] = None


class V1ListMeta(ResourceValue):
    remainingItemCount: int = 0
    resourceVersion: Optional[str] = None
    selfLink: Optional[str] = None


class ItemList(UserList, Generic[R]):
    metadata: V1ListMeta = Field(default_factory=V1ListMeta)

    def __init__(self, initlist, metadata):
        self.metadata = pydantic.parse_obj_as(V1ListMeta, metadata)
        super().__init__(initlist)

    def pop(self) -> R:
        return super().pop()

    def __getitem__(self, index) -> R:
        return super().__getitem__(index)

    def __iter__(self) -> Iterator[R]:
        return super().__iter__()


def get_default(name: str):
    from . import all

    return getattr(all, name)()


def get_type(kind: str, version: str, default: Any = None) -> Any:
    from . import all

    return all.mapping.get(kind, {}).get(version, default)



from unittest.mock import MagicMock

import pytest

from kubernetes_dynamic.kube.exceptions import NotFoundError
from kubernetes_dynamic.models.resource_item import CheckResult, ResourceItem


def test_resource_init_dict():
    item = ResourceItem(dict(key1="val1", key2={"subkey1": 3}, key3=[{"subl1": 4}]))
    assert item.key1 == "val1"
    assert item.key2.subkey1 == 3
    assert item.key3[0].subl1 == 4
    assert item.kind == ""
    assert item.apiVersion == ""


def test_resource_api(mock_client):
    """Resource API test."""
    mock_client.get_api.return_value = MagicMock(api_version="v1", kind="Kind")
    item = ResourceItem(dict(apiVersion="v1", kind="Kind", key1="val1", key2={"subkey1": 3}, key3=[{"subl1": 4}]))
    assert item.api is item.api


def test_resource_refresh(mock_client):
    item = ResourceItem(metadata={"name": "name", "namespace": "namespace"})

    mock_client.get_api.return_value.get.return_value = {"metadata": {"name": "name", "namespace": "namespace"}}
    assert item.refresh_() == item

    mock_client.get_api.return_value.get.return_value = None
    with pytest.raises(NotFoundError):
        item.refresh_()


def test_resource_is_ready(mock_client, mocker):
    is_ready_mock = mocker.patch("kubernetes_dynamic.models.resource_item.ResourceItem.check_object_is_ready")
    item = ResourceItem(metadata={"name": "name", "namespace": "namespace"})

    assert item.is_ready() == is_ready_mock.return_value
    mock_client.get_api.return_value.get.assert_not_called()
    assert item.is_ready(refresh=True) == is_ready_mock.return_value
    mock_client.get_api.return_value.get.assert_called()


def test_resource_delete(mock_client):
    item = ResourceItem(metadata={"name": "name", "namespace": "namespace"})
    assert item.delete_() == mock_client.get_api.return_value.delete.return_value
    mock_client.get_api.return_value.delete.assert_called()


@pytest.mark.parametrize(
    "conditions, result",
    [
        ([], CheckResult(False, "No conditions found on Kind name.")),
        (
            [{"status": "False"}, {"status": "True"}],
            CheckResult(False, "Condition not true on Kind name : ResourceValue:\n  status: 'False'\n."),
        ),
        (
            [{"status": "True"}, {"status": "True"}],
            CheckResult(True, "All conditions true on Kind name."),
        ),
    ],
)
def test_check_object_conditions(conditions, result):
    item = ResourceItem(
        dict(
            kind="Kind",
            metadata={"name": "name", "namespace": "namespace"},
            status={"conditions": conditions},
        )
    )
    res = ResourceItem.check_object_conditions(item)
    assert res == result
    assert bool(res) == result.state
    assert res == result.state


@pytest.mark.parametrize(
    "status, result",
    [
        ({}, CheckResult(True, "All 0 replicas ready on Kind name.")),
        (
            {"readyReplicas": 0, "replicas": 0},
            CheckResult(True, "All 0 replicas ready on Kind name."),
        ),
        (
            {"readyReplicas": 0, "replicas": 1},
            CheckResult(
                False,
                "Kind name only have 0 ready " "replicas out of required 1.",
            ),
        ),
        (
            {"readyReplicas": 1, "replicas": 1},
            CheckResult(True, "All 1 replicas ready on Kind name."),
        ),
    ],
)
def test_check_replicas_ready(status, result):
    item = ResourceItem(
        dict(
            kind="Kind",
            metadata={"name": "name", "namespace": "namespace"},
            status=status,
        )
    )
    res = ResourceItem.check_replicas_ready(item)
    assert res == result
    assert bool(res) == result.state
    assert res == result.state


def test_check_object_is_ready_no():
    item = MagicMock(status=None)
    assert ResourceItem.check_object_is_ready(item) == CheckResult(False, "Object has no status description.")


def test_check_object_is_ready_replicas(mocker):
    mock_replicas = mocker.patch("kubernetes_dynamic.models.resource_item.ResourceItem.check_replicas_ready")
    item = ResourceItem(
        dict(
            kind="Kind",
            metadata={"name": "name", "namespace": "namespace"},
            status={"replicas": 1},
        )
    )
    assert ResourceItem.check_object_is_ready(item) == mock_replicas.return_value


def test_check_object_is_ready_conditions(mocker):
    mock_conditions = mocker.patch("kubernetes_dynamic.models.resource_item.ResourceItem.check_object_conditions")
    item = ResourceItem(
        dict(
            kind="Kind",
            metadata={"name": "name", "namespace": "namespace"},
            status={"conditions": {"status": "True"}},
        )
    )
    assert ResourceItem.check_object_is_ready(item) == mock_conditions.return_value



from typing import Any
from unittest.mock import MagicMock

import pytest

from kubernetes_dynamic._kubernetes import ResourceApi
from kubernetes_dynamic.client import K8sClient
from kubernetes_dynamic.exceptions import InvalidParameter
from kubernetes_dynamic.formatters import format_selector
from kubernetes_dynamic.kube.exceptions import (
    ConfigException,
    ResourceNotUniqueError,
)
from kubernetes_dynamic.models.pod import V1Pod


def resource_api(namespaced=True, obj_type=None):
    return MagicMock(namespaced=namespaced, _resource_type=obj_type)


def fake_pod():
    return V1Pod.parse_obj({"kind": "Pod", "apiVersion": "v1"})


def fake_pods():
    return [V1Pod.parse_obj({"kind": "Pod", "apiVersion": "v1", "metadata": {"name": "pod-name"}})]


def test_k8s_client_init_incluster(mock_load_incluster: MagicMock):
    mock_load_incluster.side_effect = None
    cl = K8sClient()
    assert cl.in_cluster


def test_k8s_client_init_not_incluster(mock_load_incluster: MagicMock):
    mock_load_incluster.side_effect = ConfigException
    cl = K8sClient()
    assert not cl.in_cluster


def test_k8s_find_context():
    loader = MagicMock()
    contexts = loader.list_contexts
    contexts.return_value = [
        [],
        {"notcontext": {"cluster": "cluster_name_0"}},
        {"name": "context_name_1", "context": {"cluster": "cluster_name_1"}},
        {"name": "context_name_2", "context": {"cluster": "cluster_name_2"}},
    ]
    assert K8sClient.find_context(None, loader) is None
    assert K8sClient.find_context("cluster_name_1", loader) == "context_name_1"
    assert K8sClient.find_context("context_name_2", loader) == "context_name_2"
    with pytest.raises(RuntimeError):
        K8sClient.find_context("cluster_name", loader)


@pytest.mark.parametrize(
    "name, object_type, api_version, kind, filter",
    [
        ("pods", V1Pod, "v1", "Pod", {}),
        (None, V1Pod, "v1", "Pod", {}),
        (None, V1Pod, None, "Pod", {}),
        (None, V1Pod, "v1", None, {}),
        (None, None, "v1", "Pod", {}),
        ("pods", None, "v1", None, {}),
        ("pods", None, "v1", None, {"something": "value"}),
    ],
)
def test_k8s_client_get_api(
    name: str,
    object_type: Any,
    api_version: str,
    kind: str,
    filter: dict,
    mock_resources: MagicMock,
):
    mock_resources.get.return_value = MagicMock(kind="Pod", api_version="v1")
    cl = K8sClient()
    api = cl.get_api(name=name, object_type=object_type, api_version=api_version, kind=kind, **filter)
    filter_dict = filter
    if api_version:
        filter_dict["api_version"] = api_version
    if kind:
        filter_dict["kind"] = kind
    if name:
        filter_dict["name"] = name
    assert api == mock_resources.get.return_value
    assert api.resource_type == V1Pod
    mock_resources.get.assert_called_with(**filter_dict)


@pytest.mark.parametrize(
    "name, object_type, api_version, kind, filter",
    [
        ("pods", V1Pod, "v1", "Pod", {}),
        (None, V1Pod, "v1", "Pod", {}),
        (None, V1Pod, None, "Pod", {}),
        (None, V1Pod, "v1", None, {}),
        (None, None, "v1", "Pod", {}),
        ("pods", None, "v1", None, {}),
    ],
)
def test_k8s_client_get_api_not_unique(
    name: str,
    object_type: Any,
    api_version: str,
    kind: str,
    filter: dict,
    mock_resources: MagicMock,
):
    mock_resources.get.side_effect = ResourceNotUniqueError
    mock_resources.search.return_value = [
        MagicMock(kind="Pod", api_version="v2", preferred=False),
        MagicMock(kind="Pod", api_version="v1", preferred=True),
        MagicMock(kind="Pod", api_version="v1", preferred=True, spec=ResourceApi),
        MagicMock(kind="Pod", api_version="v4", preferred=False),
    ]
    cl = K8sClient()
    api = cl.get_api(name=name, object_type=object_type, api_version=api_version, kind=kind, **filter)
    filter_dict = filter
    if api_version:
        filter_dict["api_version"] = api_version
    if kind:
        filter_dict["kind"] = kind
    if name:
        filter_dict["name"] = name
    assert api == mock_resources.search.return_value[2]
    assert api.resource_type == V1Pod
    mock_resources.search.assert_called_with(**filter_dict)
    mock_resources.get.assert_called_with(**filter_dict)


def test_k8s_client_getattr(mock_resources: MagicMock):
    mock_resources.get.return_value = MagicMock(kind="Pod", api_version="v1")
    cl = K8sClient()
    assert cl.pods == mock_resources.get.return_value

    with pytest.raises(AttributeError):
        cl.__notexisting


@pytest.mark.parametrize(
    "selector, output",
    [
        ["string", "string"],
        [["label1", "label2"], "label1,label2"],
        [("label1", "label2"), "label1,label2"],
        [{"key": "string"}, "key=string"],
        [{"!key": "string"}, "key!=string"],
        [{"key": ["val1", "val2"]}, "key in (val1,val2)"],
        [{"!key": ["val1", "val2"]}, "key notin (val1,val2)"],
        [{"key": True}, "key"],
        [{"!key": None}, "!key"],
        [{"!k1": None, "k2": True, "k3": ["v1", "v2"]}, "!k1,k2,k3 in (v1,v2)"],
    ],
)
def test_k8s_client_format_selector(selector, output: str):
    """Test label selector formatting."""
    assert format_selector(selector) == output


def test_k8s_client_stream():
    method = MagicMock()
    method.return_value.read_all.return_value = "data"
    assert K8sClient().stream(method, "name", "namespace") == "data"


@pytest.mark.parametrize(
    "namespace, file_path, data, error, exc",
    [
        ("namespace", "file/path", {}, "mutually exclusive", InvalidParameter),
        ("namespace", "file/path", {"data": ""}, "mutually exclusive", InvalidParameter),
        ("namespace", None, None, "must be provided", InvalidParameter),
        ("namespace", "file/path", None, "file/path", FileNotFoundError),
    ],
)
def test_k8s_client_apply_errors(namespace, file_path, data, error, exc, mocker):
    path_mock = mocker.patch("kubernetes_dynamic.client.Path")
    path_mock.return_value.exists.return_value = False
    with pytest.raises(exc, match=error):
        K8sClient().apply(namespace=namespace, file_path=file_path, data=data)


def test_k8s_client_apply_empty(mocker):
    path_mock = mocker.patch("kubernetes_dynamic.client.Path")
    path_mock.return_value.exists.return_value = True
    mocker.patch("kubernetes_dynamic.client.open")
    yaml_mock = mocker.patch("kubernetes_dynamic.client.yaml")
    yaml_mock.full_load_all.return_value = []
    assert K8sClient().apply(namespace="namespace", file_path="file/path") == []


def test_k8s_client_apply(mocker, mock_resources):
    pod_api = MagicMock(kind="Pod", api_version="v1")
    mock_resources.get.return_value = pod_api
    path_mock = mocker.patch("kubernetes_dynamic.client.Path")
    path_mock.return_value.exists.return_value = True
    mocker.patch("kubernetes_dynamic.client.open")
    yaml_mock = mocker.patch("kubernetes_dynamic.client.yaml")
    yaml_mock.full_load_all.return_value = []
    item = {"kind": "Pod", "apiVersion": "v1", "metadata": {"name": "pod-name"}}
    expected = MagicMock()

    pod_api.apply.return_value = expected

    cl = K8sClient()
    cl.config.namespace = "my_namespace"
    assert cl.apply(data=item) == [expected]



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Tuple, cast

from pydantic import Field

from kubernetes_dynamic.formatters import format_selector
from kubernetes_dynamic.models.common import ItemList, V1ObjectMeta, get_default
from kubernetes_dynamic.models.resource_item import ResourceItem

if TYPE_CHECKING:
    from .all import V1ServiceSpec, V1ServiceStatus
    from .pod import V1Pod


class V1Service(ResourceItem):
    metadata: V1ObjectMeta = Field(default_factory=lambda: V1ObjectMeta())
    spec: V1ServiceSpec = Field(default_factory=lambda: get_default("V1ServiceSpec"))
    status: V1ServiceStatus = Field(default_factory=lambda: get_default("V1ServiceStatus"))

    def get_pods(self) -> ItemList[V1Pod]:
        label_selector = self.spec.selector
        return self._client.pods.get(label_selector=format_selector(label_selector))

    def get_access(self, port: str | int) -> Optional[Tuple[str, str, int]]:
        for pod in self.get_pods():
            if pod.status.phase != "Running":
                continue
            port_number = pod.get_port(port)
            if not port_number:
                continue
            return (
                pod.metadata.name,
                cast(str, pod.status.podIP),
                port_number,
            )
        return None



from datetime import datetime

from kubernetes_dynamic.models.pod import V1Pod


def test_pod_init():
    pod = V1Pod.parse_obj(dict(metadata={"name": "my_pod"}))
    assert pod.metadata.name == "my_pod"


def test_pod_get_restarts(mock_client):
    finished_at = datetime.utcnow()
    pod_1 = V1Pod.parse_obj(
        dict(
            metadata={"name": "pod-1"},
            status={
                "containerStatuses": [
                    {
                        "name": "container-1",
                        "restartCount": 1,
                        "lastState": {"terminated": {"reason": "reason", "finishedAt": finished_at}},
                    }
                ]
            },
        )
    )
    pod_2 = V1Pod.parse_obj(
        dict(
            metadata={"name": "pod-2"},
            status={
                "containerStatuses": [
                    {
                        "name": "container-2",
                        "restartCount": 0,
                        "lastState": {"terminated": {"reason": "", "finishedAt": None}},
                    }
                ]
            },
        )
    )

    mock_client.pods.get.return_value = [pod_1, pod_2]
    assert V1Pod.get_restarts() == {
        "pod-1": {
            "container-1": {
                "last_restart_finished_at": finished_at,
                "last_restart_reason": "reason",
                "restart_count": 1,
            }
        },
        "pod-2": {
            "container-2": {
                "last_restart_finished_at": "N/A",
                "last_restart_reason": "N/A",
                "restart_count": 0,
            }
        },
    }


def test_pod_exec(mock_client):
    pod = V1Pod.parse_obj(dict(metadata={"name": "my_pod", "namespace": "my_namespace"}))
    assert pod.exec("command", "container") == mock_client.stream.return_value
    mock_client.stream.assert_called_with(
        pod._api.exec.get,
        "my_pod",
        "my_namespace",
        container="container",
        command="command",
        stdin=True,
        stdout=True,
        stderr=True,
        tty=True,
    )


def test_pod_disk_usage(mock_client):
    pod = V1Pod.parse_obj(dict(metadata={"name": "my_pod"}))
    mock_client.stream.return_value = """    Used 1K-blocks    Avail Use% Mounted on
12370080 104845292 92475212  12% /
        0     65536    65536   0% /dev
        0  16440724 16440724   0% /sys/fs/cgroup
12370080 104845292 92475212  12% /tmp
        0     65536    65536   0% /dev/shm
        8  32022312 32022304   1% /opt/kx/lic
        12  32022312 32022300   1% /run/secrets/kubernetes.io/serviceaccount
        0  16440724 16440724   0% /proc/acpi
        0  16440724 16440724   0% /proc/scsi
        0  16440724 16440724   0% /sys/firmware"""
    assert pod.disk_usage("container") == {
        "/": {"used": 12370080, "size": 104845292, "avail": 92475212, "pcent": 12},
        "/dev": {"used": 0, "size": 65536, "avail": 65536, "pcent": 0},
        "/sys/fs/cgroup": {"used": 0, "size": 16440724, "avail": 16440724, "pcent": 0},
        "/tmp": {"used": 12370080, "size": 104845292, "avail": 92475212, "pcent": 12},
        "/dev/shm": {"used": 0, "size": 65536, "avail": 65536, "pcent": 0},
        "/opt/kx/lic": {"used": 8, "size": 32022312, "avail": 32022304, "pcent": 1},
        "/run/secrets/kubernetes.io/serviceaccount": {
            "used": 12,
            "size": 32022312,
            "avail": 32022300,
            "pcent": 1,
        },
        "/proc/acpi": {"used": 0, "size": 16440724, "avail": 16440724, "pcent": 0},
        "/proc/scsi": {"used": 0, "size": 16440724, "avail": 16440724, "pcent": 0},
        "/sys/firmware": {"used": 0, "size": 16440724, "avail": 16440724, "pcent": 0},
    }


def test_pod_get_controller_type_replicaset():
    pod = V1Pod.parse_obj(dict(metadata={"name": "my_pod", "ownerReferences": [{"name": "", "kind": "ReplicaSet"}]}))
    assert pod.get_controller_type() == "Deployment"


def test_pod_get_controller_type_other():
    pod = V1Pod.parse_obj(dict(metadata={"name": "my_pod", "ownerReferences": [{"name": "", "kind": "Other"}]}))
    assert pod.get_controller_type() == "Other"


def test_pod_get_env(mock_client):
    pod = V1Pod.parse_obj(dict(metadata={"name": "my_pod"}))
    mock_client.stream.return_value = """RT_KODBC_ASSEMBLY_NORTH_0_SERVICE_HOST=10.218.54.243
INSIGHTS_GUI_GATEWAY_PORT_10001_TCP_PROTO=tcp
RT_KODBC_ASSEMBLY_SOUTH_2_PORT_7000_UDP_PORT=7000
RT_KODBC_ASSEMBLY_NORTH_1_PORT_6000_TCP=tcp://10.218.53.117:6000
RT_KODBC_ASSEMBLY_NORTH_2_PORT_8000_UDP_ADDR=10.218.52.165
INSIGHTS_KEYCLOAK_SERVICE_HOST=10.218.52.249
_=/usr/bin/env"""
    assert pod.get_env() == {
        "RT_KODBC_ASSEMBLY_NORTH_0_SERVICE_HOST": "10.218.54.243",
        "INSIGHTS_GUI_GATEWAY_PORT_10001_TCP_PROTO": "tcp",
        "RT_KODBC_ASSEMBLY_SOUTH_2_PORT_7000_UDP_PORT": "7000",
        "RT_KODBC_ASSEMBLY_NORTH_1_PORT_6000_TCP": "tcp://10.218.53.117:6000",
        "RT_KODBC_ASSEMBLY_NORTH_2_PORT_8000_UDP_ADDR": "10.218.52.165",
        "INSIGHTS_KEYCLOAK_SERVICE_HOST": "10.218.52.249",
        "_": "/usr/bin/env",
    }



from unittest.mock import MagicMock

from kubernetes_dynamic.kube.exceptions import ConflictError
from kubernetes_dynamic.models.namespace import V1Namespace


def test_namespace_init():
    pod = V1Namespace.parse_obj(dict(metadata={"name": "my_namespace"}))
    assert pod.metadata.name == "my_namespace"


def test_namespace_ensure(mock_client):
    mock_client.get_api.return_value.create.return_value = V1Namespace.parse_obj(
        dict(metadata={"name": "my_namespace"})
    )
    namespace = V1Namespace.ensure("my_namespace")
    assert namespace.metadata.name == "my_namespace"


def test_namespace_ensure_conflict(mock_client):
    mock_client.get_api.return_value.create.side_effect = ConflictError(MagicMock(name="my_namespace"))
    mock_client.get_api.return_value.get.return_value = V1Namespace.parse_obj(dict(metadata={"name": "my_namespace"}))
    namespace = V1Namespace.ensure("my_namespace")
    assert namespace.metadata.name == "my_namespace"


def test_namespace_annotate(mock_client):
    namespace = V1Namespace.parse_obj(dict(metadata={"name": "my_namespace"}))
    mock_client.get_api.return_value.get.return_value = V1Namespace.parse_obj(
        dict(metadata={"name": "my_namespace", "annotations": {"something": "1"}})
    )
    mock_client.get_api.return_value.patch.return_value = V1Namespace.parse_obj(
        dict(metadata={"name": "my_namespace", "annotations": {"something": "1", "something_2": "2"}})
    )
    namespace.annotate({"something_2": "2"})
    assert namespace.metadata.annotations == {"something": "1", "something_2": "2"}
    mock_client.get_api.return_value.patch.assert_called()



from typing import Union

SelectorTypes = Union[None, str, dict, list, tuple]


def _format_dict_selector(selector: dict):
    items = []
    for key, value in selector.items():
        exists = True
        if key.startswith("!"):
            exists = False
            key = key[1:]
        if isinstance(value, (tuple, list)):
            items.append(f"{key} {'in' if exists else 'notin'} ({format_selector(value)})")
        elif isinstance(value, str):
            items.append(f"{key}{'!' if not exists else ''}={value}")
        elif not exists or value is False:
            items.append(f"!{key}")
        else:
            items.append(key)
    return format_selector(items)


def format_selector(selector: SelectorTypes):
    """Format kubernetes selector.

    Args:
        selector: The selector object to convert.
            - string: no conversion done.
            - list, tuple: the object is concatenated by `,`.
            - dict:
                - key: string           -> key=string             # key equals string
                - !key: string          -> key!=string            # key not equals string
                - key: [val1, val2]     -> key in (val1, val2)    # key is val1 OR val2
                - !key: [val1, val2]    -> key notin (val1, val2) # key is NOT val1 AND NOT val2
                - key: True             -> key                    # key exists
                - key: False            -> !key                   # key NOT exists
                - !key: None            -> !key                   # key NOT exists

    Returns:
        Kubernetes compatible string selector
    """
    if isinstance(selector, (list, tuple)):
        return ",".join(selector)
    if isinstance(selector, dict):
        return _format_dict_selector(selector)
    return selector


def to_camel(string: str) -> str:
    """Convert snake_case to CamelCase"""
    return "".join(word.capitalize() for word in string.split("_"))


def to_lower_camel(string: str) -> str:
    """Convert snake_case to lowerCamelCase"""
    if len(string) >= 1:
        pascal_string = to_camel(string)
        return pascal_string[0].lower() + pascal_string[1:]
    return string.lower()
